"""Interact with hive through JDBC

Reference:
- https://docs.cloudera.com/HDPDocuments/HDP3/HDP-3.1.5/integrating-hive/content/hive_connection_string_url_syntax.html
"""
from pathlib import Path
import jaydebeapi
from loguru import logger
from typing import Union
from khalinox import config, utils
from urllib.parse import urlencode, urlparse


def connect(myconfig: config.Config) -> Union[jaydebeapi.Connection, None]:
    """Interact with hive using jdbc

    Raises:
        ValueError: for bad user/password
    """
    host = urlparse(myconfig.knox_url)
    url = _make_connection_str(host.netloc, myconfig.hive_jdbc)
    try:
        return jaydebeapi.connect(
            myconfig.hive_jdbc.driver_class,
            url,
            {
                "user": myconfig.user,
                "password": utils.decrypt(myconfig._key, myconfig.password),
            },
        )

    # this kind of exception seems to come from java somehow
    # and does not raise a specific one in python
    # pylint: disable=broad-except
    except Exception as e:
        if "code: 401" in str(e):
            user = myconfig.user
            basemsg = (
                "\nCannot authenticate against Knox/HiveServer (401).\n"
                "Please check your user/password"
            )
            # common mistake, user provide `MYUSER` instead of `myuser`
            if user.isupper():
                extra = (
                    f"\nSuggestion: your login `{user}` is in upper case,"
                    f"try in lower case (`{user.lower()}`)"
                )
                msg = basemsg + extra
            else:
                msg = basemsg
            logger.critical(msg)
            raise ValueError("Bad credential") from None
        elif "code: 500" in str(e):
            logger.critical("Internal error with Hive. Most Likely due to Hive outage")
            raise ValueError("Hive server Error") from None
        else:
            raise e


def _make_connection_str(host: str, hiveconf: config.HiveJDBC) -> str:
    """Build JDBC connection string
    `jdbc:hive2://<host>:<port>/<dbName>;<sessionConfs>?<hiveConfs>#<hiveVars>`

    Args:
        host: include port,e.g `myhost.com:8884`
    """
    hive_vars = urlencode(hiveconf.hive_vars)
    return (
        f"jdbc:hive2://{host}/;"
        f"ssl=true;sslTrustStore={hiveconf.trust_store};"
        f"trustStorePassword={hiveconf.store_password};"
        f"transportMode=http;httpPath=gateway/default/hive?{hive_vars}"
    )



"""helpers
"""

from cryptography.fernet import Fernet
from toolz import pipe


def _tostr(data: bytes) -> str:
    return data.decode("utf8")


def _tobytes(data: str) -> bytes:
    return bytes(data, "utf8")


def encrypt(key: str, password: str) -> str:
    f = Fernet(_tobytes(key))
    return pipe(password, _tobytes, f.encrypt, _tostr)


def decrypt(key: str, encrypted: str) -> str:
    f = Fernet(_tobytes(key))
    return pipe(encrypted, _tobytes, f.decrypt, _tostr)



"""pythonic hdfs client
"""
import requests

# from hdfs import InsecureClient,Client
import hdfs
from khalinox import config, utils
from urllib.parse import urljoin

import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def _webhdfs_url(knox_url: str, path="gateway/default/webhdfs/v1") -> str:
    # return "https://obitedhs-vcs001.equant.com:8443/gateway/default/webhdfs/v1"
    return urljoin(knox_url, path)


def hdfs_client(myconfig: config.Config) -> hdfs.InsecureClient:
    # context to use with knox simple user/password auth
    session = requests.Session()
    session.verify = myconfig.verify
    # https://docs.python-requests.org/en/master/user/advanced/#session-objects
    session.auth = (myconfig.user, utils.decrypt(myconfig._key, myconfig.password))
    return hdfs.InsecureClient(url=_webhdfs_url(myconfig.knox_url), session=session)



"""Store and retrieve credentials and configuration

Handle serialization, deserialization and encryption for password.
"""

import getpass
import json
import os
from pathlib import Path
from typing import Callable

import requests
import urllib3
from loguru import logger
from urllib3.exceptions import InsecureRequestWarning
from pydantic import BaseSettings, HttpUrl, validator
from pathlib import Path
from typing import Union, TYPE_CHECKING
from khalinox import utils


# hacky for autocomplete in vscode
# https://github.com/microsoft/python-language-server/issues/1898

if TYPE_CHECKING:  # pragma: no cover
    import dataclasses

    static_check_init_args = dataclasses.dataclass
else:

    def static_check_init_args(cls):
        return cls


@static_check_init_args
class WebHDFS(BaseSettings):
    # where to write/read by default
    default_dir: Path


@static_check_init_args
class HiveJDBC(BaseSettings):
    trust_store: Path
    store_password: str
    hive_vars: dict
    driver_class: str = "org.apache.hive.jdbc.HiveDriver"


@static_check_init_args
class Config(BaseSettings):
    """Configuration to interact with hive & hdfs using knox

    user/password are asked using interactive inputs if they are not provided
    """

    knox_url: HttpUrl
    # passed verbatim to python requests
    verify: Union[str, bool]
    user: str = ""
    # encrypt/decrypt
    _key: str
    # encrypted with key
    password: str = ""
    path: Path
    webhdfs: WebHDFS
    hive_jdbc: HiveJDBC

    def __init__(self, **data):
        super().__init__(**data)
        # this could also be done with default_factory
        try:
            self._key = os.environ["KNOX_KEY"]
        except KeyError as k:
            raise KeyError("fernet key should exist in env as `KNOX_KEY`")

    @validator("knox_url")
    def port_must_be_present(cls, url):
        if not url.port:
            raise ValueError(
                f"{url} should provide a specific port (e.g. https://myhost.me:8888)"
            )
        return url

    @validator("password")
    def encrypt_password(cls, password):
        if password:
            return utils.encrypt(os.environ["KNOX_KEY"], password)
        else:
            raw = getpass.getpass("password")
            return utils.encrypt(os.environ["KNOX_KEY"], raw)

    @validator("user")
    def non_empty_user(cls, user):
        if user:
            return user
        else:
            raw = input("user")
            return raw

    class Config:
        env_prefix = "knox_"
        underscore_attrs_are_private = True

    def save(self):
        self.path.write_text(self.json(), encoding="utf8")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['khalinox']

package_data = \
{'': ['*']}

install_requires = \
['JayDeBeApi>=1.2.3,<2.0.0',
 'cryptography>=3.4.7,<4.0.0',
 'hdfs>=2.6.0,<3.0.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.8.1,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'toolz>=0.11.1,<0.12.0']

setup_kwargs = {
    'name': 'khalinox',
    'version': '1.1.0',
    'description': '',
    'long_description': None,
    'author': 'Khalid',
    'author_email': 'khalid.chakhmoun@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
